from torch import softmax
from torch.distributions import Categorical
import torch
from tqdm.notebook import tqdm 
from transformers import AutoModelForCausalLM, AutoTokenizer



def next_token_distr(model, ids):
    logits = model(ids.unsqueeze(0))['logits'][0][-1]
    probs = softmax(logits, dim=0)
    return probs


def generate_next_token(model, ids, return_distr=False):
    probs = next_token_distr(model, ids)
    sample = Categorical(probs).sample()
    return sample if not return_distr else sample, probs


def generate_tokens(model: AutoModelForCausalLM, tokenizer: AutoTokenizer, ids, stop_ids, max_length=50, return_distr=False):
    i = 0
    answer = []
    next_token = -1

    bar = tqdm(total=max_length)

    if return_distr:
        distr = []
    
    with torch.no_grad():
        while i < max_length and next_token != tokenizer.eos_token_id and not torch.all(torch.eq(ids[-len(stop_ids):], stop_ids)):
            next_token = generate_next_token(model, ids, return_distr=return_distr)
            
            if return_distr:
                next_token, next_distr = next_token
                distr.append(next_distr)

            ids = torch.cat((ids,next_token.unsqueeze(0)), dim=0)
            answer.append(next_token)
            bar.update()
            i += 1
        while i < max_length:
            bar.update()
            i += 1

    if return_distr:    
        distr = torch.stack(distr)
    answer = torch.stack(answer)

    return answer if not return_distr else answer, distr





from transformers import AutoModelForCausalLM, AutoTokenizer
from rethink.utils import generate_tokens
import torch


default_prompt = "{}\nH: {}\nM: "
default_stop = "H:"


def process_context(
    request,
    context,
    rethinking_prompt,
    tokenizer
):
    rethinking_request = rethinking_prompt.format(
        context,
        request
    )
    ids = tokenizer.encode(rethinking_request, return_tensors='pt')[0]
    return ids


def correct_answer(
        model: AutoModelForCausalLM, 
        tokenizer: AutoTokenizer, 
        request,
        context,
        rethinking_prompt=default_prompt, 
        max_length=50,
        stop_word=default_stop,
        return_distr=False
    ):
    # generate prompt for context-aware generation
    ids = process_context(request, context, rethinking_prompt, tokenizer).to(model.device)
    stop_ids = tokenizer.encode(stop_word, return_tensors='pt')[0].to(model.device)
    
    new_tokens = generate_tokens(model, tokenizer, ids, stop_ids, max_length, return_distr)
    if return_distr:
        answer, distr = new_tokens

    if torch.all(answer[-len(stop_ids):] == stop_ids):

        answer = answer[:-len(stop_ids)]
    answer = tokenizer.decode(answer, skip_special_tokens=True)
    return answer if not return_distr else answer, distr



from setuptools import setup, find_packages

setup(
    name='rethink',
    version='0.0.2',
    packages=find_packages(include=['rethink', 'rethink.*']),
    install_requires=[
        'torch',
        'transformers',
        'tqdm',
    ]
)